@@ -7,7 +7,7 @@ use Carp ();
 use IO::Handle ();
 use Exporter 'import';
 
-our $VERSION = '0.010';
+our $VERSION = '0.011';
 
 use constant DEFAULT_REQUEST_BODY_LIMIT => 16777216;
 use constant DEFAULT_REQUEST_BODY_BUFFER => 262144;
@@ -310,17 +310,19 @@ sub _body_params {
           require Encode;
           $name = Encode::decode($default_charset, "$name");
         }
-        if (!defined $headers->{'content-type'} or $headers->{'content-type'} =~ m/^text\/plain\b/i) {
-          my $value_charset = $default_charset;
-          if (defined $headers->{'content-type'}) {
-            if (my ($charset_quoted, $charset_unquoted) = $headers->{'content-type'} =~ m/;\s*charset=(?:"((?:\\[\\"]|[^"])+)"|([^";]+))/i) {
-              $charset_quoted =~ s/\\([\\"])/$1/g if defined $charset_quoted;
-              $value_charset = defined $charset_quoted ? $charset_quoted : $charset_unquoted;
-            }
+        my $value_charset;
+        if (defined $headers->{'content-type'}) {
+          if (my ($charset_quoted, $charset_unquoted) = $headers->{'content-type'} =~ m/;\s*charset=(?:"((?:\\[\\"]|[^"])+)"|([^";]+))/i) {
+            $charset_quoted =~ s/\\([\\"])/$1/g if defined $charset_quoted;
+            $value_charset = defined $charset_quoted ? $charset_quoted : $charset_unquoted;
           }
-          if (length $value_charset) {
-            require Encode;
+        }
+        if (defined $value_charset or !defined $headers->{'content-type'} or $headers->{'content-type'} =~ m/^text\/plain\b/i) {
+          require Encode;
+          if (defined $value_charset) {
             $value = Encode::decode($value_charset, "$value");
+          } elsif (length $default_charset) {
+            $value = Encode::decode($default_charset, "$value");
           }
         }
         push @names, $name unless exists $keyed{$name};
@@ -440,6 +442,8 @@ sub set_nph {
   return $self;
 }
 
+sub set_response_fixed_length { $_[0]{response_fixed_length} = $_[1]; $_[0] }
+
 sub set_response_body_buffer { $_[0]{response_body_buffer} = $_[1]; $_[0] }
 
 sub set_response_status {
@@ -459,26 +463,29 @@ sub set_response_status {
   return $self;
 }
 
-sub set_response_content_disposition {
-  my ($self, $disposition, $filename) = @_;
-  if ($self->{headers_rendered}) {
-    Carp::carp "Attempted to set HTTP response content disposition but headers have already been rendered";
-  } else {
-    Carp::croak "Invalid characters in HTTP response content disposition" if defined $disposition and $disposition =~ m/[^a-zA-Z0-9!#\$%&'*+\-.^_`|~]/;
-    $self->{response_content_disposition} = $disposition;
-    # filename will be quoted/escaped later
-    $self->{response_filename} = $filename;
+{
+  my %DISPOSITIONS = (attachment => 1, inline => 1);
+  sub set_response_disposition {
+    my ($self, $disposition, $filename) = @_;
+    if ($self->{headers_rendered}) {
+      Carp::carp "Attempted to set HTTP response content disposition but headers have already been rendered";
+    } else {
+      Carp::croak "Attempted to set unknown Content-Disposition value '$disposition'" unless exists $DISPOSITIONS{lc $disposition};
+      $self->{response_disposition} = $disposition;
+      # filename will be quoted/escaped later
+      $self->{response_filename} = $filename;
+    }
+    return $self;
   }
-  return $self;
 }
 
-sub set_response_content_type {
+sub set_response_type {
   my ($self, $content_type) = @_;
   if ($self->{headers_rendered}) {
     Carp::carp "Attempted to set HTTP response content type but headers have already been rendered";
   } else {
     Carp::croak "Newline characters not allowed in HTTP response content type" if defined $content_type and $content_type =~ tr/\r\n//;
-    $self->{response_content_type} = $content_type;
+    $self->{response_type} = $content_type;
   }
   return $self;
 }
@@ -550,18 +557,45 @@ sub headers_rendered { $_[0]{headers_rendered} }
     Carp::croak "Don't know how to render '$type'" if length $type and !exists $RENDER_TYPES{$type};
     my $charset = $self->{response_charset};
     $charset = 'UTF-8' unless defined $charset;
+
+    my $fixed_length = $self->{response_fixed_length} && !$self->{headers_rendered};
+    my ($response_body, $response_length);
+    if ($type eq 'json') {
+      $response_body = $self->_json->encode($data);
+      $response_length = length $response_body if $fixed_length;
+    } elsif ($type eq 'html' or $type eq 'xml' or $type eq 'text') {
+      if (uc $charset eq 'UTF-8' and do { local $@; eval { require Unicode::UTF8; 1 } }) {
+        $response_body = Unicode::UTF8::encode_utf8($data);
+      } else {
+        require Encode;
+        $response_body = Encode::encode($charset, "$data");
+      }
+      $response_length = length $response_body if $fixed_length;
+    } elsif ($type eq 'data') {
+      $response_body = $data;
+      $response_length = length $response_body if $fixed_length;
+    } elsif ($type eq 'file' and $fixed_length) {
+      $response_length = -s $data;
+      Carp::croak "Failed to retrieve size of file '$data': $!" unless defined $response_length;
+    } elsif ($type eq 'redirect' or ($fixed_length and !length $type)) {
+      $response_length = 0;
+    }
+
     my $out_fh = defined $self->{output_handle} ? $self->{output_handle} : *STDOUT;
     if (!$self->{headers_rendered}) {
-      my @headers = @{$self->{response_headers} || []};
       my $headers_str = '';
       my %headers_set;
-      foreach my $header (@headers) {
+      foreach my $header (@{$self->{response_headers} || []}) {
         my ($name, $value) = @$header;
         $headers_str .= "$name: $value\r\n";
         $headers_set{lc $name} = 1;
       }
-      if (!$headers_set{'content-disposition'} and (defined $self->{response_content_disposition} or defined $self->{response_filename})) {
-        my $value = defined $self->{response_content_disposition} ? $self->{response_content_disposition} : 'inline';
+      if (!$headers_set{'content-length'} and defined $response_length) {
+        $headers_str = "Content-Length: $response_length\r\n$headers_str";
+        $self->{response_fixed_length_set} = 1;
+      }
+      if (!$headers_set{'content-disposition'} and (defined $self->{response_disposition} or defined $self->{response_filename})) {
+        my $value = defined $self->{response_disposition} ? $self->{response_disposition} : 'inline';
         if (defined(my $filename = $self->{response_filename})) {
           require Encode;
           my $quoted_filename = Encode::encode('ISO-8859-1', "$filename");
@@ -579,7 +613,7 @@ sub headers_rendered { $_[0]{headers_rendered} }
         $headers_str = "Location: $data\r\n$headers_str";
       }
       if (!$headers_set{'content-type'} and $type ne 'redirect') {
-        my $content_type = $self->{response_content_type};
+        my $content_type = $self->{response_type};
         $content_type =
             $type eq 'json' ? 'application/json;charset=UTF-8'
           : $type eq 'html' ? "text/html;charset=$charset"
@@ -610,18 +644,13 @@ sub headers_rendered { $_[0]{headers_rendered} }
       $self->{headers_rendered} = 1;
     } elsif ($type eq 'redirect') {
       Carp::carp "Attempted to render a redirect but headers have already been rendered";
+    } elsif ($self->{response_fixed_length_set}) {
+      Carp::carp "Attempted to render additional response data but a fixed length response has already been rendered";
+      return $self;
     }
-    if ($type eq 'json') {
-      $out_fh->printflush($self->_json->encode($data));
-    } elsif ($type eq 'html' or $type eq 'xml' or $type eq 'text') {
-      if (uc $charset eq 'UTF-8' and do { local $@; eval { require Unicode::UTF8; 1 } }) {
-        $out_fh->printflush(Unicode::UTF8::encode_utf8($data));
-      } else {
-        require Encode;
-        $out_fh->printflush(Encode::encode($charset, "$data"));
-      }
-    } elsif ($type eq 'data') {
-      $out_fh->printflush($data);
+
+    if ($type eq 'json' or $type eq 'html' or $type eq 'xml' or $type eq 'text' or $type eq 'data') {
+      $out_fh->printflush($response_body);
     } elsif ($type eq 'file' or $type eq 'handle') {
       my $in_fh;
       if ($type eq 'file') {
@@ -636,6 +665,7 @@ sub headers_rendered { $_[0]{headers_rendered} }
       }
       $out_fh->flush;
     }
+    return $self;
   }
 }
 
